#   '$Name:  $pp`';
#   '$Header:  $';
# =============================================================================
#
# file :       device.py
#
# description : Device server to control the Keithley Multimeters 2000 serie
# from Keithley Instruments, Inc.
#
# project :     TANGO Device Server
#
# $Author:  Roger Pastor Ortiz
#
# $Revision:  $
#
# $Log:  $
#
# copyleft :    Alba Synchrotron Radiation Facility
#               Cerdanyola del Valles
#               Spain
#
#
#         (c) - Controls group - ALBA
# =============================================================================

import PyTango
from keithley20xx import Keithley, KeithleyError
import threading
import numpy as np
import time


DEV_STATE_UNKNOWN = PyTango.DevState.UNKNOWN
DEV_STATE_FAULT = PyTango.DevState.FAULT
DEV_STATE_RUNNING = PyTango.DevState.RUNNING
DEV_STATE_ON = PyTango.DevState.ON
DEV_STATE_INIT = PyTango.DevState.INIT
DEV_STATE_STANDBY = PyTango.DevState.STANDBY
DEV_STATE_MOVING = PyTango.DevState.MOVING
DEV_STATE_ALARM = PyTango.DevState.ALARM


def ExceptionHandler(func):
    """
        Class to decorate the exception management
        :param func: function
    """
    def new_func(*args, **kwargs):
        """
        Method to execute function an input an entry on debug stream and
        warning in case of failure
        :param kwargs: function params
        :return: value: if it is a possible one
        """
        try:
            obj = args[0]
            obj.debug_stream('Entering in %s (%s)' % (func.__name__,
                                                      repr(args)))
            result = func(*args, **kwargs)
            obj.debug_stream('Exiting %s' % func.__name__)
            return result
        except Exception as e:
            obj.warn_stream('Hardware warning in %s: %s' % (func.__name__, e))
            raise e
    return new_func


class KeithleyBaseClass(PyTango.DeviceClass):
    """
        Keithley multimeter base class
    """
    # Class Properties
    class_property_list = {
        }

    # Device Properties
    device_property_list = {
        'GPIB_DS_Path': [PyTango.DevString, 'Keithley GPIB DS path',
                         'test/rpastor/gpib_1'],
        'Logfile': [PyTango.DevString, 'Keithley logfile path',
                    '/tmp/DS/Keithley.log'],
        'PowerFrequency': [PyTango.DevDouble, 'Keithley power feed '
                                              'frequency', '50'],
        }

    # Command definitions
    cmd_list = {
                'ClearStatus': [[PyTango.DevVoid, 'Clears all event registers '
                                'and Error Queue and puts the Keithley on '
                                'idle state.'], [PyTango.DevVoid, 'None']],

                'Reset': [[PyTango.DevVoid, 'Returns model to reset '
                                            'conditions'],
                          [PyTango.DevVoid, 'None']],

                # HW commands
                # 'FunctionMeasurement': [[PyTango.DevString, 'VOLT'],
                #                         [PyTango.DevString, 'None']],


                # 'CurrentAcquire': [[PyTango.DevVoid, 'None'],
                #                    [PyTango.DevString, 'None']],
                # 'VoltageAcquire': [[PyTango.DevVoid, 'None'],
                #                    [PyTango.DevString, 'None']],
                'Preset': [[PyTango.DevVoid, 'None'],
                           [PyTango.DevVoid, 'None']],
                'ClearErrorQueue': [[PyTango.DevVoid, 'None'],
                                    [PyTango.DevVoid, 'None']],
                'Start': [[PyTango.DevVoid, 'None'],
                             [PyTango.DevVoid, 'None']],
                'Stop': [[PyTango.DevVoid, 'None'],
                          [PyTango.DevVoid, 'None']],
                # 'TriggerBus': [[PyTango.DevVoid, 'None'],
                #                [PyTango.DevVoid, 'None']],
                # 'Wait': [[PyTango.DevVoid, 'None'],
                #          [PyTango.DevVoid, 'None']],

                # 'NonBlockingReadConfig': [[PyTango.DevVoid, 'None'],
                #                           [PyTango.DevVoid, 'None']],

                # 'MakeAcquisition': [[PyTango.DevShort, 'Make <n> acquisitions '
                #                                        'with the actual '
                #                                        'function'],
                #                     [PyTango.DevVarStringArray, 'None']],
                'PrepareAcquisition': [[PyTango.DevVoid, 'None'],
                                       [PyTango.DevVoid, 'None']],
                'RawDataIndex': [[PyTango.DevVarShortArray, 'None'],
                                 [PyTango.DevVarFloatArray, 'None']],
                # 'RawCmdRead': [[PyTango.DevVoid, 'None'],
                #                [PyTango.DevString, 'Read string']],
                'RawCmdWrite': [[PyTango.DevString, 'Write string'],
                                [PyTango.DevVoid, 'None']],
                'RawCmdWriteRead': [[PyTango.DevString, 'Write string'],
                                    [PyTango.DevString, 'Read string']],

    }

    # Attributes list
    attr_list = {
        'Measure': [[PyTango.ArgType.DevString,
                     PyTango.AttrDataFormat.SCALAR,
                     PyTango.AttrWriteType.READ],
                    {'description': 'Make a post-processed reading'}],

        'Calculate1Function': [[PyTango.ArgType.DevString,
                                PyTango.AttrDataFormat.SCALAR,
                                PyTango.AttrWriteType.READ_WRITE]],
        'Calculate1Enable': [[PyTango.ArgType.DevString,
                              PyTango.AttrDataFormat.SCALAR,
                              PyTango.AttrWriteType.READ_WRITE]],
        'Calculate1Data': [[PyTango.ArgType.DevString,
                            PyTango.AttrDataFormat.SCALAR,
                            PyTango.AttrWriteType.READ]],
        'Calculate2Function': [[PyTango.ArgType.DevString,
                                PyTango.AttrDataFormat.SCALAR,
                                PyTango.AttrWriteType.READ_WRITE]],
        'Calculate2Enable': [[PyTango.ArgType.DevString,
                              PyTango.AttrDataFormat.SCALAR,
                              PyTango.AttrWriteType.READ_WRITE]],
        'Calculate2Data': [[PyTango.ArgType.DevString,
                            PyTango.AttrDataFormat.SCALAR,
                            PyTango.AttrWriteType.READ]],

        'OutputDataType': [[PyTango.ArgType.DevString,
                           PyTango.AttrDataFormat.SCALAR,
                           PyTango.AttrWriteType.READ_WRITE],
                           {'memorized': 'true', }],
        'OutputDataElements': [[PyTango.ArgType.DevString,
                               PyTango.AttrDataFormat.SPECTRUM,
                               PyTango.AttrWriteType.READ_WRITE, 6]],
                               # {'memorized': 'true', }],
        'OutputBinaryOrder': [[PyTango.ArgType.DevString,
                              PyTango.AttrDataFormat.SCALAR,
                              PyTango.AttrWriteType.READ_WRITE],
                              {'memorized': 'true', }],

        'ReadLastMeasure': [[PyTango.ArgType.DevString,
                             PyTango.AttrDataFormat.SCALAR,
                             PyTango.AttrWriteType.READ]],

        'CurrentIntegrationTime': [[PyTango.ArgType.DevDouble,
                                   PyTango.AttrDataFormat.SCALAR,
                                   PyTango.AttrWriteType.READ_WRITE],
                                   {'memorized': 'true', }],
        'VoltageIntegrationTime': [[PyTango.ArgType.DevDouble,
                                   PyTango.AttrDataFormat.SCALAR,
                                   PyTango.AttrWriteType.READ_WRITE],
                                   {'memorized': 'true', }],

        'CurrentRange': [[PyTango.ArgType.DevDouble,
                          PyTango.AttrDataFormat.SCALAR,
                          PyTango.AttrWriteType.READ_WRITE]],
        'VoltageRange': [[PyTango.ArgType.DevDouble,
                          PyTango.AttrDataFormat.SCALAR,
                          PyTango.AttrWriteType.READ_WRITE]],
        'CurrentAutoRange': [[PyTango.ArgType.DevBoolean,
                             PyTango.AttrDataFormat.SCALAR,
                             PyTango.AttrWriteType.READ_WRITE],
                             {'memorized': 'true', }],
        'VoltageAutoRange': [[PyTango.ArgType.DevBoolean,
                             PyTango.AttrDataFormat.SCALAR,
                             PyTango.AttrWriteType.READ_WRITE],
                             {'memorized': 'true', }],

        'CurrentDigits': [[PyTango.ArgType.DevShort,
                          PyTango.AttrDataFormat.SCALAR,
                          PyTango.AttrWriteType.READ_WRITE],
                          {'memorized': 'true', }],
        'VoltageDigits': [[PyTango.ArgType.DevShort,
                          PyTango.AttrDataFormat.SCALAR,
                          PyTango.AttrWriteType.READ_WRITE],
                          {'memorized': 'true', }],

        'VoltageInRange': [[PyTango.ArgType.DevDouble,
                            PyTango.AttrDataFormat.SCALAR,
                            PyTango.AttrWriteType.READ_WRITE]],
        # 'CurrentInThreshold': [[PyTango.ArgType.DevDouble,
        #                         PyTango.AttrDataFormat.SCALAR,
        #                         PyTango.AttrWriteType.READ_WRITE]],
        # 'VoltageInThreshold': [[PyTango.ArgType.DevDouble,
        #                         PyTango.AttrDataFormat.SCALAR,
        #                         PyTango.AttrWriteType.READ_WRITE]],


        'MeasEventRegister': [[PyTango.ArgType.DevShort,
                               PyTango.AttrDataFormat.SCALAR,
                               PyTango.AttrWriteType.READ]],
        'MeasEventRegisterEnable': [[PyTango.ArgType.DevShort,
                                    PyTango.AttrDataFormat.SCALAR,
                                    PyTango.AttrWriteType.READ_WRITE],
                                    {'memorized': 'true', }],

        'PowerOnState': [[PyTango.ArgType.DevString,
                         PyTango.AttrDataFormat.SCALAR,
                         PyTango.AttrWriteType.READ_WRITE]],
        'ErrorMsg': [[PyTango.ArgType.DevString,
                     PyTango.AttrDataFormat.SCALAR,
                     PyTango.AttrWriteType.READ]],

        'AutoZeroEnable': [[PyTango.ArgType.DevBoolean,
                           PyTango.AttrDataFormat.SCALAR,
                           PyTango.AttrWriteType.READ_WRITE],
                           {'memorized': 'true', }],


        'InitiateContinuous': [[PyTango.ArgType.DevBoolean,
                               PyTango.AttrDataFormat.SCALAR,
                               PyTango.AttrWriteType.READ_WRITE],
                               {'memorized': 'true', }],


        'CountMeasureLayer': [[PyTango.ArgType.DevDouble,
                              PyTango.AttrDataFormat.SCALAR,
                              PyTango.AttrWriteType.READ_WRITE]],


        'DelayMeasureLayer': [[PyTango.ArgType.DevDouble,
                              PyTango.AttrDataFormat.SCALAR,
                              PyTango.AttrWriteType.READ_WRITE]],

        'SourceMeasureLayer': [[PyTango.ArgType.DevString,
                               PyTango.AttrDataFormat.SCALAR,
                               PyTango.AttrWriteType.READ_WRITE]],


        'TimerMeasureLayer': [[PyTango.ArgType.DevDouble,
                              PyTango.AttrDataFormat.SCALAR,
                              PyTango.AttrWriteType.READ_WRITE]],

        'EventEnableRegister': [[PyTango.ArgType.DevShort,
                                PyTango.AttrDataFormat.SCALAR,
                                PyTango.AttrWriteType.READ_WRITE],
                                {'memorized': 'true', }],
        'EventStatusRegister': [[PyTango.ArgType.DevShort,
                                PyTango.AttrDataFormat.SCALAR,
                                PyTango.AttrWriteType.READ]],

        'Identification': [[PyTango.ArgType.DevString,
                           PyTango.AttrDataFormat.SCALAR,
                           PyTango.AttrWriteType.READ]],

        'OperationCompleted': [[PyTango.ArgType.DevBoolean,
                               PyTango.AttrDataFormat.SCALAR,
                               PyTango.AttrWriteType.READ_WRITE]],
        'MemoryOptionInstalled': [[PyTango.ArgType.DevString,
                                  PyTango.AttrDataFormat.SCALAR,
                                  PyTango.AttrWriteType.READ]],

        'ServiceRequestEnableRegister': [[PyTango.ArgType.DevShort,
                                         PyTango.AttrDataFormat.SCALAR,
                                         PyTango.AttrWriteType.READ_WRITE],
                                         {'memorized': 'true', }],
        'StatusByteRegister': [[PyTango.ArgType.DevShort,
                               PyTango.AttrDataFormat.SCALAR,
                               PyTango.AttrWriteType.READ_WRITE],
                               {'memorized': 'true', }],

        'ErrorFlag': [[PyTango.ArgType.DevBoolean,
                      PyTango.AttrDataFormat.SCALAR,
                      PyTango.AttrWriteType.READ]],
        # 'ReadNonBlocking': [[PyTango.ArgType.DevString,
        #                      PyTango.AttrDataFormat.SCALAR,
        #                      PyTango.AttrWriteType.READ]],
        'AcqModeEnable': [[PyTango.ArgType.DevBoolean,
                          PyTango.AttrDataFormat.SCALAR,
                          PyTango.AttrWriteType.READ_WRITE]],
        'RawData': [[PyTango.ArgType.DevString,
                    PyTango.AttrDataFormat.SCALAR,
                    PyTango.AttrWriteType.READ]],
        'RawDataBuffer': [[PyTango.ArgType.DevString,
                          PyTango.AttrDataFormat.SPECTRUM,
                          PyTango.AttrWriteType.READ, 9999]],
        'OperationEventRegister': [[PyTango.ArgType.DevBoolean,
                                   PyTango.AttrDataFormat.SCALAR,
                                   PyTango.AttrWriteType.READ]],
        'TimerPeriod': [[PyTango.ArgType.DevDouble,
                        PyTango.AttrDataFormat.SCALAR,
                        PyTango.AttrWriteType.READ_WRITE]],
        'ChunkSize': [[PyTango.ArgType.DevShort,
                      PyTango.AttrDataFormat.SCALAR,
                      PyTango.AttrWriteType.READ_WRITE],
                      {'memorized': 'true', }],
        'DataReady': [[PyTango.ArgType.DevShort,
                       PyTango.AttrDataFormat.SCALAR,
                       PyTango.AttrWriteType.READ]],

    }

    def __init__(self, name):
        """
        Method to initialise a device as Keithley
        :return:
        """
        PyTango.DeviceClass.__init__(self, name)
        self.set_type("KeithleyDevice")

class Keithley2001Class(KeithleyBaseClass):
    """
    Specific API for Keithley 2001 model.
    """
    add_cmd_list = {
        'SaveSetup': [[PyTango.DevVoid, 'None'], [PyTango.DevVoid, 'None']],
        'LoadSetup': [[PyTango.DevVoid, 'None'], [PyTango.DevVoid, 'None']],
        'ClearBuffer': [[PyTango.DevVoid, 'None'], [PyTango.DevVoid, 'None']],

    }

    add_attr_list = {
        'MeasurementFunction': [[PyTango.ArgType.DevString,
                                 PyTango.AttrDataFormat.SCALAR,
                                 PyTango.AttrWriteType.READ_WRITE],
                                {'description': 'Make a one-shot measure of '
                                                'the given function'}],
        'CurrentAutoIntegrationTime': [[PyTango.ArgType.DevBoolean,
                                        PyTango.AttrDataFormat.SCALAR,
                                        PyTango.AttrWriteType.READ_WRITE]],
        'VoltageAutoIntegrationTime': [[PyTango.ArgType.DevBoolean,
                                        PyTango.AttrDataFormat.SCALAR,
                                        PyTango.AttrWriteType.READ_WRITE]],

        'CurrentCoupled': [[PyTango.ArgType.DevString,
                            PyTango.AttrDataFormat.SCALAR,
                            PyTango.AttrWriteType.READ_WRITE]],
        'VoltageCoupled': [[PyTango.ArgType.DevString,
                            PyTango.AttrDataFormat.SCALAR,
                            PyTango.AttrWriteType.READ_WRITE]],

        'CurrentAutoDigits': [[PyTango.ArgType.DevBoolean,
                               PyTango.AttrDataFormat.SCALAR,
                               PyTango.AttrWriteType.READ_WRITE],
                              {'memorized': 'true', }],
        'VoltageAutoDigits': [[PyTango.ArgType.DevBoolean,
                               PyTango.AttrDataFormat.SCALAR,
                               PyTango.AttrWriteType.READ_WRITE],
                               {'memorized': 'true', }],

        'InputSource': [[PyTango.ArgType.DevString,
                         PyTango.AttrDataFormat.SCALAR,
                         PyTango.AttrWriteType.READ_WRITE]],
        'CurrentInRange': [[PyTango.ArgType.DevDouble,
                            PyTango.AttrDataFormat.SCALAR,
                            PyTango.AttrWriteType.READ_WRITE]],

        'DigitalInPortData': [[PyTango.ArgType.DevBoolean,
                               PyTango.AttrDataFormat.SCALAR,
                               PyTango.AttrWriteType.READ]],

        'AutoZero': [[PyTango.ArgType.DevString,
                      PyTango.AttrDataFormat.SCALAR,
                      PyTango.AttrWriteType.READ_WRITE],
                      {'memorized': 'true', }],

        'AcquisitionMethod': [[PyTango.ArgType.DevString,
                               PyTango.AttrDataFormat.SCALAR,
                               PyTango.AttrWriteType.READ_WRITE]],

        'LineSynchronization': [[PyTango.ArgType.DevBoolean,
                                 PyTango.AttrDataFormat.SCALAR,
                                 PyTango.AttrWriteType.READ_WRITE]],


        'CountArmLayer': [[PyTango.ArgType.DevDouble,
                           PyTango.AttrDataFormat.SCALAR,
                           PyTango.AttrWriteType.READ_WRITE]],
        'CountScanLayer': [[PyTango.ArgType.DevDouble,
                            PyTango.AttrDataFormat.SCALAR,
                            PyTango.AttrWriteType.READ_WRITE]],

        'DelayScanLayer': [[PyTango.ArgType.DevDouble,
                            PyTango.AttrDataFormat.SCALAR,
                            PyTango.AttrWriteType.READ_WRITE]],

        'SourceArmLayer': [[PyTango.ArgType.DevString,
                           PyTango.AttrDataFormat.SCALAR,
                           PyTango.AttrWriteType.READ_WRITE]],
        'SourceScanLayer': [[PyTango.ArgType.DevString,
                            PyTango.AttrDataFormat.SCALAR,
                            PyTango.AttrWriteType.READ_WRITE]],

        'TimerScanLayer': [[PyTango.ArgType.DevDouble,
                            PyTango.AttrDataFormat.SCALAR,
                            PyTango.AttrWriteType.READ_WRITE]],

        'CoupleTriggers': [[PyTango.ArgType.DevString,
                            PyTango.AttrDataFormat.SCALAR,
                            PyTango.AttrWriteType.READ_WRITE]],

        'InlineArmControl': [[PyTango.ArgType.DevShort,
                             PyTango.AttrDataFormat.SCALAR,
                             PyTango.AttrWriteType.READ_WRITE],
                             {'memorized': 'true', }],
        'InlineScanControl': [[PyTango.ArgType.DevShort,
                              PyTango.AttrDataFormat.SCALAR,
                              PyTango.AttrWriteType.READ_WRITE],
                              {'memorized': 'true', }],
        'InlineMeasureControl': [[PyTango.ArgType.DevShort,
                                 PyTango.AttrDataFormat.SCALAR,
                                 PyTango.AttrWriteType.READ_WRITE],
                                 {'memorized': 'true', }],
        'OutlineArmControl': [[PyTango.ArgType.DevShort,
                              PyTango.AttrDataFormat.SCALAR,
                              PyTango.AttrWriteType.READ_WRITE],
                              {'memorized': 'true', }],
        'OutlineScanControl': [[PyTango.ArgType.DevShort,
                               PyTango.AttrDataFormat.SCALAR,
                               PyTango.AttrWriteType.READ_WRITE],
                               {'memorized': 'true', }],
        'OutlineMeasureControl': [[PyTango.ArgType.DevShort,
                                  PyTango.AttrDataFormat.SCALAR,
                                  PyTango.AttrWriteType.READ_WRITE],
                                  {'memorized': 'true', }],

        'SemiSynchronousLine': [[PyTango.ArgType.DevShort,
                                 PyTango.AttrDataFormat.SCALAR,
                                 PyTango.AttrWriteType.READ_WRITE]],
        'TriggerEventRegister': [[PyTango.ArgType.DevBoolean,
                                 PyTango.AttrDataFormat.SCALAR,
                                 PyTango.AttrWriteType.READ]],

        # Buffer associated cmd (only for k2001 as it has 30k positions buffer)
        'BufferElementGroup': [[PyTango.ArgType.DevString,
                               PyTango.AttrDataFormat.SCALAR,
                               PyTango.AttrWriteType.READ_WRITE]],
        'BufferSize': [[PyTango.ArgType.DevShort,
                       PyTango.AttrDataFormat.SCALAR,
                       PyTango.AttrWriteType.READ_WRITE]],
        'BufferSizeAuto': [[PyTango.ArgType.DevBoolean,
                           PyTango.AttrDataFormat.SCALAR,
                           PyTango.AttrWriteType.READ_WRITE]],
        'BufferFeed': [[PyTango.ArgType.DevString,
                       PyTango.AttrDataFormat.SCALAR,
                       PyTango.AttrWriteType.READ_WRITE]],
        'BufferControl': [[PyTango.ArgType.DevString,
                          PyTango.AttrDataFormat.SCALAR,
                          PyTango.AttrWriteType.READ_WRITE]],
        'BufferData': [[PyTango.ArgType.DevString,
                       PyTango.AttrDataFormat.SPECTRUM,
                       PyTango.AttrWriteType.READ, 30000]],
        'BufferDataStored': [[PyTango.ArgType.DevShort,
                             PyTango.AttrDataFormat.SCALAR,
                             PyTango.AttrWriteType.READ]],



    }

    def __init__(self, name):
        KeithleyBaseClass.__init__(self, name)
        self.cmd_list = self.cmd_list.copy()
        self.attr_list = self.attr_list.copy()
        self.cmd_list.update(self.add_cmd_list)
        self.attr_list.update(self.add_attr_list)


class KeithleyDevice(PyTango.Device_4Impl):
    # ------------------------------------------------------------------
    #   Device constructor
    # ------------------------------------------------------------------
    def __init__(self, cl, name):
        """
        Method to decorate Keithley device object instantiation and to set the
        state machine
        :param cl:
        :param name:
        """
        PyTango.Device_4Impl.__init__(self, cl, name)
        self.info_stream('In KeithleyDevice.__init__')
        try:
            self.init_device()
        except Exception as e:
            self.set_state(DEV_STATE_FAULT)
            self.error_stream('Hardware error in __init__(): %s' % e)
            raise e
        self.state_machine()

    # ------------------------------------------------------------------
    #   Device destructor
    # ------------------------------------------------------------------
    def delete_device(self):
        """
        Method to destroy a Keithley device object.
        """
        self.keithley.__del__()
        self.info_stream('In %s::delete_device()' % self.get_name())

    # ------------------------------------------------------------------
    #   Device initialization
    # ------------------------------------------------------------------
    def init_device(self):
        """
        Method to initiate a Keithley object.
        """
        self.info_stream('In %s::init_device()' % self.get_name())
        klass = self.get_device_class()
        # Gets the properties to init the device instance
        self.get_device_properties(klass)

        state = DEV_STATE_INIT
        status = 'Keithley device initiated.'

        self.init_failure = False
        try:
            self.keithley = Keithley(gpib_ds_path=self.GPIB_DS_Path,
                                     logfile=self.Logfile)
            self.keithley.error_msg_clear()
        except:
            status = 'Failure on communication with the HW (i.e. GPIB DS)'
            state = DEV_STATE_UNKNOWN
            self.init_failure = True
        # Initialize attributes
        self.async = False

        # Set a flag to indicate that the server fires change events manually
        # without the polling to be started for the attribute
        # Define events on attributes
        self.set_change_event('Measure', True, False)
        self.set_change_event('Calculate1Function', True, False)
        self.set_change_event('Calculate1Enable', True, False)
        self.set_change_event('Calculate1Data', True, False)
        self.set_change_event('Calculate2Function', True, False)
        self.set_change_event('Calculate2Enable', True, False)
        self.set_change_event('Calculate2Data', True, False)
        self.set_change_event('OutputDataType', True, False)
        self.set_change_event('OutputDataElements', True, False)
        self.set_change_event('OutputBinaryOrder', True, False)
        self.set_change_event('CurrentIntegrationTime', True, False)
        self.set_change_event('VoltageIntegrationTime', True, False)
        self.set_change_event('CurrentRange', True, False)
        self.set_change_event('VoltageRange', True, False)
        self.set_change_event('CurrentAutoRange', True, False)
        self.set_change_event('VoltageAutoRange', True, False)
        self.set_change_event('CurrentDigits', True, False)
        self.set_change_event('VoltageDigits', True, False)
        self.set_change_event('VoltageInRange', True, False)
        # self.set_change_event('CurrentInThreshold', True, False)
        # self.set_change_event('VoltageInThreshold', True, False)
        self.set_change_event('MeasEventRegisterEnable', True, False)
        self.set_change_event('PowerOnState', True, False)
        self.set_change_event('AutoZeroEnable', True, False)
        self.set_change_event('InitiateContinuous', True, False)
        self.set_change_event('CountMeasureLayer', True, False)
        self.set_change_event('DelayMeasureLayer', True, False)
        self.set_change_event('SourceMeasureLayer', True, False)
        self.set_change_event('TimerMeasureLayer', True, False)
        self.set_change_event('EventEnableRegister', True, False)
        self.set_change_event('OperationCompleted', True, False)
        self.set_change_event('ServiceRequestEnableRegister', True, False)
        self.set_change_event('StatusByteRegister', True, False)
        # self.set_change_event('ReadNonBlocking', True, False)
        self.set_change_event('AcqModeEnable', True, False)
        self.set_change_event('TimerPeriod', True, False)

        self.get_device_attr().get_attr_by_name('DataReady')
        self.attr_raw_data = self.get_device_attr().get_attr_by_name(
                'DataReady')
        self.attr_raw_data.set_data_ready_event(True)
        # print(dir(self.attr_raw_data))
        # self.set_change_event('RawData', True, False)



        self.set_change_event('State', True, False)
        self.set_change_event('Status', True, False)

        if isinstance(klass, Keithley2001Class):
            self.set_change_event('MeasurementFunction', True, False)
            self.set_change_event('TimerScanLayer', True, False)
            self.set_change_event('CoupleTriggers', True, False)
            self.set_change_event('InlineArmControl', True, False)
            self.set_change_event('InlineScanControl', True, False)
            self.set_change_event('SourceArmLayer', True, False)
            self.set_change_event('SourceScanLayer', True, False)
            self.set_change_event('DelayScanLayer', True, False)
            self.set_change_event('CurrentAutoIntegrationTime', True, False)
            self.set_change_event('VoltageAutoIntegrationTime', True, False)
            self.set_change_event('CurrentCoupled', True, False)
            self.set_change_event('VoltageCoupled', True, False)
            self.set_change_event('CurrentAutoDigits', True, False)
            self.set_change_event('VoltageAutoDigits', True, False)
            self.set_change_event('InputSource', True, False)
            self.set_change_event('CurrentInRange', True, False)
            self.set_change_event('AutoZero', True, False)
            self.set_change_event('AcquisitionMethod', True, False)
            self.set_change_event('LineSynchronization', True, False)
            self.set_change_event('CountArmLayer', True, False)
            self.set_change_event('CountScanLayer', True, False)
            self.set_change_event('InlineMeasureControl', True, False)
            self.set_change_event('OutlineArmControl', True, False)
            self.set_change_event('OutlineScanControl', True, False)
            self.set_change_event('OutlineMeasureControl', True, False)
            self.set_change_event('SemiSynchronousLine', True, False)

            self.set_change_event('BufferElementGroup', True, False)
            self.set_change_event('BufferSize', True, False)
            self.set_change_event('BufferSizeAuto', True, False)
            self.set_change_event('BufferFeed', True, False)
            self.set_change_event('BufferControl', True, False)
            self.set_change_event('BufferData', True, False)
            self.set_change_event('BufferDataStored', True, False)

        self.set_state(state)
        self.set_status(status)
        self.push_change_event('State', self.get_state())
        self.push_change_event('Status', self.get_status())

        self.quit_event = threading.Event()
        self.acq_mode = False
        self.raw_data = []
        self.error_list = []

        # self.Stop()
        self.idle = True

        self.data_ready_chunk = 8



        self.count_index = 0

        self.dyn_attr()

    # ------------------------------------------------------------------
    #   Create dynamic attributes
    # ------------------------------------------------------------------

    def dyn_attr(self):
        """
        Method to create attributes dynamically
        """
        pass

    # ------------------------------------------------------------------
    #   State machine implementation
    # ------------------------------------------------------------------
    @ExceptionHandler
    def state_machine(self):
        """
        Method to manage the state machine
        :return: None
        """
        if self.async:
            return

        if self.init_failure:
            return

        if not self.idle:
            self.idle = (self.keithley.operation_event_register & 2**10 != 0)

        if self.keithley.error_flag or self.error_list != []:

            while self.keithley.error_flag:
                self.error_list += [self.keithley.error_msg]
                print('\n\n\nDEBUG: error_list is... \n' + str(self.error_list))


            if '-211,"Trigger ignored"' in self.error_list:
                status = 'WARNING: Keithley trigger lost'
                state = DEV_STATE_ALARM


            if self.error_list.count('-211,"Trigger ignored"') != \
                    len(self.error_list):

                state = DEV_STATE_FAULT
                status = 'Keithley on fault. List of errors: '
                for error in self.error_list:
                    status += '\n' + str(error)
                status += '\n' + 'Execute ClearErrorQueue cmd to get out of ' \
                                 'FAULT state.'
        elif self.idle:
            status = 'Keithley on idle state'
            state = DEV_STATE_ON

        elif not self.idle:
            state = DEV_STATE_RUNNING

            if not self.acq_mode:
                status = 'Keithley taking measures'
            else:
                status = 'Keithley acquiring'

        else:
            state = DEV_STATE_UNKNOWN

        self.set_state(state)
        self.set_status(status)

    # ------------------------------------------------------------------
    #   Always executed hook method
    # ------------------------------------------------------------------
    def always_executed_hook(self):
        """
        Tango method to execute always before any method that logs stream
        info, updates the state machine and the State and Status
        :return: None
        """
        self.info_stream('In %s::always_executed_hook()' % self.get_name())
        self.state_machine()
        self.push_change_event('State', self.get_state())
        self.push_change_event('Status', self.get_status())

    # ------------------------------------------------------------------
    #   ATTRIBUTES
    # ------------------------------------------------------------------
    def read_attr_hardware(self, data):
        pass

    # ------------------------------------------------------------------
    #   read & write MeasurementFunction
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_MeasurementFunction(self, the_att):
        """
        Method to read the MeasurementFunction and update its value.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.configure)

    @ExceptionHandler
    def write_MeasurementFunction(self, the_att):
        """
        Method to set the MeasurementFunction
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.configure = data[0]
        self.push_change_event('MeasurementFunction', data[0])

    def is_MeasurementFunction_allowed(self, req_type):
        """
        Method to check if MeasurementFunction is available to
        interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_ON, DEV_STATE_INIT, )

    # ------------------------------------------------------------------
    #   read Measure
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_Measure(self, the_att):
        """
        Method to read the Measured value
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.read_postprocessed)

    def is_Measure_allowed(self, req_type):
        """
        Method to check if the Measure is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_ON, DEV_STATE_INIT, )

    # ------------------------------------------------------------------
    #   read & write Calculate1Function
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_Calculate1Function(self, the_att):
        """
        Method to read the Calculate1Function value
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.calculate_1_function)

    @ExceptionHandler
    def write_Calculate1Function(self, the_att):
        """
        Method to set the Calculate1Function
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.calculate_1_function = data[0]
        self.push_change_event('Calculate1Function', data[0])

    def is_Calculate1Function_allowed(self, req_type):
        """
        Method to check if the Calculate1Function is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_ON, DEV_STATE_INIT, )

    # ------------------------------------------------------------------
    #   read & write Calculate1Enable
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_Calculate1Enable(self, the_att):
        """
        Method to read the Calculate1Enable value
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.calculate_1_enable)

    @ExceptionHandler
    def write_Calculate1Enable(self, the_att):
        """
        Method to set the Calculate1Enable
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.calculate_1_enable = data[0]
        self.push_change_event('Calculate1Enable', data[0])

    def is_Calculate1Enable_allowed(self, req_type):
        """
        Method to check if the Calculate1Enable is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_ON, DEV_STATE_INIT, )

    # ------------------------------------------------------------------
    #   read Calculate1Data
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_Calculate1Data(self, the_att):
        """
        Method to read the Calculate1Data value
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.calculate_1_data)

    def is_Calculate1Data_allowed(self, req_type):
        """
        Method to check if the Calculate1Data is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_ON, DEV_STATE_INIT,
                                    DEV_STATE_RUNNING, DEV_STATE_FAULT,)

    # ------------------------------------------------------------------
    #   read & write Calculate2Function
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_Calculate2Function(self, the_att):
        """
        Method to read the Calculate2Function value
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.calculate_2_function)

    @ExceptionHandler
    def write_Calculate2Function(self, the_att):
        """
        Method to set the Calculate2Function
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.calculate_2_function = data[0]
        self.push_change_event('Calculate2Function', data[0])

    def is_Calculate2Function_allowed(self, req_type):
        """
        Method to check if the Calculate2Function is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_ON, DEV_STATE_INIT, )

    # ------------------------------------------------------------------
    #   read & write Calculate2Enable
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_Calculate2Enable(self, the_att):
        """
        Method to read the Calculate2Enable value
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.calculate_2_enable)

    @ExceptionHandler
    def write_Calculate2Enable(self, the_att):
        """
        Method to set the Calculate2Enable
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.calculate_2_enable = data[0]
        self.push_change_event('Calculate2Enable', data[0])

    def is_Calculate2Enable_allowed(self, req_type):
        """
        Method to check if the Calculate2Enable is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_ON, DEV_STATE_INIT, )

    # ------------------------------------------------------------------
    #   read Calculate2Data
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_Calculate2Data(self, the_att):
        """
        Method to read the Calculate2Data value
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.calculate_2_data)

    def is_Calculate2Data_allowed(self, req_type):
        """
        Method to check if the Calculate2Data is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_ON, DEV_STATE_INIT,
                                    DEV_STATE_RUNNING, DEV_STATE_FAULT,)

    # ------------------------------------------------------------------
    #   read & write OutputDataType
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_OutputDataType(self, the_att):
        """
        Method to read the OutputDataType value
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.output_data_type)

    @ExceptionHandler
    def write_OutputDataType(self, the_att):
        """
        Method to set the OutputDataType
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.output_data_type = data[0]
        self.push_change_event('OutputDataType', data[0])

    def is_OutputDataType_allowed(self, req_type):
        """
        Method to check if the OutputDataType is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_ON, DEV_STATE_INIT, )

    # ------------------------------------------------------------------
    #   read & write OutputDataElements
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_OutputDataElements(self, the_att):
        """
        Method to read the OutputDataElements value
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.output_data_elements.split(','))

    @ExceptionHandler
    def write_OutputDataElements(self, the_att):
        """
        Method to set the OutputDataElements
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.output_data_elements = data
        self.push_change_event('OutputDataElements', data[0])

    def is_OutputDataElements_allowed(self, req_type):
        """
        Method to check if the OutputDataElements is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_ON, DEV_STATE_INIT, )

    # ------------------------------------------------------------------
    #   read & write OutputBinaryOrder
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_OutputBinaryOrder(self, the_att):
        """
        Method to read the OutputBinaryOrder value
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.output_data_binary_order)

    @ExceptionHandler
    def write_OutputBinaryOrder(self, the_att):
        """
        Method to set the OutputBinaryOrder
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.output_data_binary_order = data[0]
        self.push_change_event('OutputBinaryOrder', data[0])


    def is_OutputBinaryOrder_allowed(self, req_type):
        """
        Method to check if the OutputBinaryOrder is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_ON, DEV_STATE_INIT, )

    # ------------------------------------------------------------------
    #   read & write ReadLastMeasure
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_ReadLastMeasure(self, the_att):
        """
        Method to read the ReadLastMeasure value
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.read_latest)

    def is_ReadLastMeasure_allowed(self, req_type):
        """
        Method to check if the ReadLastMeasure is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_ON, DEV_STATE_RUNNING, )

    # ------------------------------------------------------------------
    #   read & write CurrentIntegrationTime
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_CurrentIntegrationTime(self, the_att):
        """
        Method to read the CurrentIntegrationTime value
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(eval(self.keithley.current_nplc)*1.0 /
                          eval(self.PowerFrequency))

    @ExceptionHandler
    def write_CurrentIntegrationTime(self, the_att):
        """
        Method to set the CurrentIntegrationTime
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.current_nplc = data[0] * 1.0 * eval(self.PowerFrequency)
        self.push_change_event('CurrentIntegrationTime', data[0])

    def is_CurrentIntegrationTime_allowed(self, req_type):
        """
        Method to check if the CurrentIntegrationTime is available to
        interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_ON, DEV_STATE_INIT, )

    # ------------------------------------------------------------------
    #   read & write VoltageIntegrationTime
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_VoltageIntegrationTime(self, the_att):
        """
        Method to read the VoltageIntegrationTime value
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(eval(self.keithley.voltage_nplc)*1.0 /
                          eval(self.PowerFrequency))

    @ExceptionHandler
    def write_VoltageIntegrationTime(self, the_att):
        """
        Method to set the VoltageIntegrationTime
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.voltage_nplc = data[0] * 1.0 * eval(self.PowerFrequency)
        self.push_change_event('VoltageIntegrationTime', data[0])

    def is_VoltageIntegrationTime_allowed(self, req_type):
        """
        Method to check if the CurrentIntegrationTime is available to
        interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_ON, DEV_STATE_INIT, )

    # ------------------------------------------------------------------
    #   read & write CurrentAutoIntegrationTime
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_CurrentAutoIntegrationTime(self, the_att):
        """
        Method to read the CurrentAutoIntegrationTime value
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(eval(self.keithley.current_auto_nplc))

    @ExceptionHandler
    def write_CurrentAutoIntegrationTime(self, the_att):
        """
        Method to set the CurrentAutoIntegrationTime
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        if data[0]:
            self.keithley.current_auto_nplc = 'ON'
        if not data[0]:
            self.keithley.current_auto_nplc = 'OFF'
        self.push_change_event('CurrentAutoIntegrationTime', data[0])

    def is_CurrentAutoIntegrationTime_allowed(self, req_type):
        """
        Method to check if the VoltageAutoIntegrationTime is available to
        interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_ON, DEV_STATE_INIT, )

    # ------------------------------------------------------------------
    #   read & write VoltageAutoIntegrationTime
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_VoltageAutoIntegrationTime(self, the_att):
        """
        Method to read the VoltageAutoIntegrationTime value
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(eval(self.keithley.voltage_auto_nplc))

    @ExceptionHandler
    def write_VoltageAutoIntegrationTime(self, the_att):
        """
        Method to set the VoltageAutoIntegrationTime
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        if data[0]:
            self.keithley.voltage_auto_nplc = 'ON'
        if not data[0]:
            self.keithley.voltage_auto_nplc = 'OFF'
        self.push_change_event('VoltageAutoIntegrationTime', data[0])

    def is_VoltageAutoIntegrationTime_allowed(self, req_type):
        """
        Method to check if the VoltageAutoIntegrationTime is available to
        interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_ON, DEV_STATE_INIT, )

    # ------------------------------------------------------------------
    #   read & write CurrentCoupled
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_CurrentCoupled(self, the_att):
        """
        Method to read the CurrentCoupled value
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.coupled_current_ac)

    @ExceptionHandler
    def write_CurrentCoupled(self, the_att):
        """
        Method to set the CurrentCoupled
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.coupled_current_ac = data[0]
        self.push_change_event('CurrentCoupled', data[0])

    def is_CurrentCoupled_allowed(self, req_type):
        """
        Method to check if the CurrentCoupled is available to
        interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_ON, DEV_STATE_INIT, )

    # ------------------------------------------------------------------
    #   read & write VoltageCoupled
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_VoltageCoupled(self, the_att):
        """
        Method to read the VoltageCoupled value
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.coupled_voltage_ac)

    @ExceptionHandler
    def write_VoltageCoupled(self, the_att):
        """
        Method to set the VoltageCoupled
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.coupled_voltage_ac = data[0]
        self.push_change_event('VoltageCoupled', data[0])

    def is_VoltageCoupled_allowed(self, req_type):
        """
        Method to check if the VoltageCoupled is available to
        interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_ON, DEV_STATE_INIT, )

    # ------------------------------------------------------------------
    #   read & write CurrentRange
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_CurrentRange(self, the_att):
        """
        Method to read the CurrentRange.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(eval(self.keithley.current_rang))

    @ExceptionHandler
    def write_CurrentRange(self, the_att):
        """
        Method to set the CurrentRange
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.current_rang = data[0]
        self.push_change_event('CurrentRange', data[0])

    def is_CurrentRange_allowed(self, req_type):
        """
        Method to check if the CurrentRange is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON,)

    # ------------------------------------------------------------------
    #   read & write VoltageRange
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_VoltageRange(self, the_att):
        """
        Method to read the VoltageRange.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(eval(self.keithley.voltage_rang))

    @ExceptionHandler
    def write_VoltageRange(self, the_att):
        """
        Method to set the VoltageRange
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.voltage_rang = data[0]
        self.push_change_event('VoltageRange', data[0])

    def is_VoltageRange_allowed(self, req_type):
        """
        Method to check if the VoltageRange is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON,)

    # ------------------------------------------------------------------
    #   read & write CurrentAutoRange
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_CurrentAutoRange(self, the_att):
        """
        Method to read the CurrentAutoRange.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(eval(self.keithley.current_auto_rang))

    @ExceptionHandler
    def write_CurrentAutoRange(self, the_att):
        """
        Method to set the CurrentAutoRange
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        if data[0]:
            self.keithley.current_auto_rang = 'ON'
        if not data[0]:
            self.keithley.current_auto_rang = 'OFF'
        self.push_change_event('CurrentAutoRange', data[0])

    def is_CurrentAutoRange_allowed(self, req_type):
        """
        Method to check if the CurrentAutoRange is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON,)
    # ------------------------------------------------------------------
    #   read & write VoltageAutoRange
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_VoltageAutoRange(self, the_att):
        """
        Method to read the VoltageAutoRange.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(eval(self.keithley.voltage_auto_rang))

    @ExceptionHandler
    def write_VoltageAutoRange(self, the_att):
        """
        Method to set the VoltageAutoRange
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        if data[0]:
            self.keithley.voltage_auto_rang = 'ON'
        if not data[0]:
            self.keithley.voltage_auto_rang = 'OFF'
        self.push_change_event('VoltageAutoRange', data[0])

    def is_VoltageAutoRange_allowed(self, req_type):
        """
        Method to check if the VoltageAutoRange is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON,)

    # ------------------------------------------------------------------
    #   read & write CurrentDigits
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_CurrentDigits(self, the_att):
        """
        Method to read the CurrentDigits.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(eval(self.keithley.current_digits))

    @ExceptionHandler
    def write_CurrentDigits(self, the_att):
        """
        Method to set the CurrentDigits
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.current_digits = data[0]
        self.push_change_event('CurrentDigits', data[0])

    def is_CurrentDigits_allowed(self, req_type):
        """
        Method to check if the CurrentDigits is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON,)

    # ------------------------------------------------------------------
    #   read & write VoltageDigits
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_VoltageDigits(self, the_att):
        """
        Method to read the VoltageDigits.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(eval(self.keithley.voltage_digits))

    @ExceptionHandler
    def write_VoltageDigits(self, the_att):
        """
        Method to set the VoltageDigits
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.voltage_digits = data[0]
        self.push_change_event('VoltageDigits', data[0])

    def is_VoltageDigits_allowed(self, req_type):
        """
        Method to check if the VoltageDigits is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON,)

    # ------------------------------------------------------------------
    #   read & write CurrentAutoDigits
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_CurrentAutoDigits(self, the_att):
        """
        Method to read the CurrentAutoDigits.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.current_auto_digits)

    @ExceptionHandler
    def write_CurrentAutoDigits(self, the_att):
        """
        Method to set the CurrentAutoDigits
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        if data[0]:
            self.keithley.current_auto_digits = 'ON'
        if not data[0]:
            self.keithley.current_auto_digits = 'OFF'
        self.push_change_event('CurrentAutoDigits', data[0])

    def is_CurrentAutoDigits_allowed(self, req_type):
        """
        Method to check if the CurrentAutoDigits is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON,)

    # ------------------------------------------------------------------
    #   read & write VoltageAutoDigits
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_VoltageAutoDigits(self, the_att):
        """
        Method to read the VoltageAutoDigits.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(eval(self.keithley.voltage_auto_digits))

    @ExceptionHandler
    def write_VoltageAutoDigits(self, the_att):
        """
        Method to set the VoltageAutoDigits
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        if data[0]:
            self.keithley.voltage_auto_digits = 'ON'
        if not data[0]:
            self.keithley.voltage_auto_digits = 'OFF'
        self.push_change_event('VoltageAutoDigits', data[0])

    def is_VoltageAutoDigits_allowed(self, req_type):
        """
        Method to check if the VoltageAutoDigits is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON,)

    # ------------------------------------------------------------------
    #   read & write InputSource
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_InputSource(self, the_att):
        """
        Method to read the InputSource.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.input_source)

    @ExceptionHandler
    def write_InputSource(self, the_att):
        """
        Method to set the InputSource
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        if data[0] in ['VOLT', 'CURR']:
            self.keithley.input_source = data[0]
            self.push_change_event('InputSource', data[0])
        else:
            raise AttributeError('Input source should be VOLT or CURR')

    def is_InputSource_allowed(self, req_type):
        """
        Method to check if the InputSource is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON,)

    # ------------------------------------------------------------------
    #   read & write CurrentInRange
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_CurrentInRange(self, the_att):
        """
        Method to read the CurrentInRange.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(eval(self.keithley.current_input_range))

    @ExceptionHandler
    def write_CurrentInRange(self, the_att):
        """
        Method to set the CurrentInRange
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.current_input_range = data[0]
        self.push_change_event('CurrentInRange', data[0])

    def is_CurrentInRange_allowed(self, req_type):
        """
        Method to check if the CurrentInRange is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_RUNNING,
                                    DEV_STATE_FAULT, DEV_STATE_ON,)

    # ------------------------------------------------------------------
    #   read & write VoltageInRange
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_VoltageInRange(self, the_att):
        """
        Method to read the VoltageInRange.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(eval(self.keithley.voltage_input_range))

    @ExceptionHandler
    def write_VoltageInRange(self, the_att):
        """
        Method to set the VoltageInRange
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.voltage_input_range = data[0]
        self.push_change_event('VoltageInRange', data[0])

    def is_VoltageInRange_allowed(self, req_type):
        """
        Method to check if the VoltageInRange is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON,)

    # # ------------------------------------------------------------------
    # #   read & write CurrentInThreshold
    # # ------------------------------------------------------------------
    # @ExceptionHandler
    # def read_CurrentInThreshold(self, the_att):
    #     """
    #     Method to read the CurrentInThreshold.
    #     :param the_att: actual name of the attribute
    #     :return: None
    #     """
    #     the_att.set_value(self.keithley.current_input_threshold)
    #
    # @ExceptionHandler
    # def write_CurrentInThreshold(self, the_att):
    #     """
    #     Method to set the CurrentInThreshold
    #     :param the_att: actual name of the attribute
    #     :return: None
    #     """
    #     data = []
    #     the_att.get_write_value(data)
    #     self.keithley.current_input_threshold = data[0]
    #     self.push_change_event('CurrentInThreshold', data[0])
    #
    # def is_CurrentInThreshold_allowed(self, req_type):
    #     """
    #     Method to check if the CurrentInThreshold is available to interact with.
    #     :return: None
    #     """
    #     return self.get_state() in (DEV_STATE_INIT, DEV_STATE_RUNNING,
    #                                 DEV_STATE_FAULT, DEV_STATE_ON,)
    #
    # # ------------------------------------------------------------------
    # #   read & write VoltageInThreshold
    # # ------------------------------------------------------------------
    # @ExceptionHandler
    # def read_VoltageInThreshold(self, the_att):
    #     """
    #     Method to read the VoltageInThreshold.
    #     :param the_att: actual name of the attribute
    #     :return: None
    #     """
    #     the_att.set_value(self.keithley.voltage_input_threshold)
    #
    # @ExceptionHandler
    # def write_VoltageInThreshold(self, the_att):
    #     """
    #     Method to set the VoltageInThreshold
    #     :param the_att: actual name of the attribute
    #     :return: None
    #     """
    #     data = []
    #     the_att.get_write_value(data)
    #     self.keithley.voltage_input_threshold = data[0]
    #     self.push_change_event('VoltageInThreshold', data[0])
    #
    # def is_VoltageInThreshold_allowed(self, req_type):
    #     """
    #     Method to check if the VoltageInThreshold is available to interact
    #     with.
    #     :return: None
    #     """
    #     return self.get_state() in (DEV_STATE_INIT, DEV_STATE_RUNNING,
    #                                 DEV_STATE_FAULT, DEV_STATE_ON,)

    # ------------------------------------------------------------------
    #   read DigitalInPortData
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_DigitalInPortData(self, the_att):
        """
        Method to read the DigitalInPortData.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(eval(self.keithley.digital_input_port))

    def is_DigitalInPortData_allowed(self, req_type):
        """
        Method to check if the DigitalInPortData is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON,)

    # ------------------------------------------------------------------
    #   read MeasEventRegister
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_MeasEventRegister(self, the_att):
        """
        Method to read the MeasEventRegister.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(eval(self.keithley.measurement_event_register))

    def is_MeasEventRegister_allowed(self, req_type):
        """
        Method to check if the MeasEventRegister is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON,)

    # ------------------------------------------------------------------
    #   read & write MeasEventRegisterEnable
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_MeasEventRegisterEnable(self, the_att):
        """
        Method to read the MeasEventRegisterEnable.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(eval(self.keithley.measurement_event_register_enable))

    @ExceptionHandler
    def write_MeasEventRegisterEnable(self, the_att):
        """
        Method to set the MeasEventRegisterEnable
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.measurement_event_register_enable = data[0]
        self.push_change_event('MeasEventRegisterEnable', data[0])

    def is_MeasEventRegisterEnable_allowed(self, req_type):
        """
        Method to check if the MeasEventRegisterEnable is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT,  DEV_STATE_ON,)

    # ------------------------------------------------------------------
    #   read & write PowerOnState
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_PowerOnState(self, the_att):
        """
        Method to read the PowerOnState.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.power_on_state)

    @ExceptionHandler
    def write_PowerOnState(self, the_att):
        """
        Method to set the PowerOnState
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.power_on_state = data[0]
        self.push_change_event('PowerOnState', data[0])

    def is_PowerOnState_allowed(self, req_type):
        """
        Method to check if the PowerOnState is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON,)

    # ------------------------------------------------------------------
    #   read & write ErrorMsg
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_ErrorMsg(self, the_att):
        """
        Method to read the ErrorMsg.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.error_msg)

    def is_ErrorMsg_allowed(self, req_type):
        """
        Method to check if the ErrorMsg is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_RUNNING,
                                    DEV_STATE_FAULT, DEV_STATE_ON,
                                    DEV_STATE_MOVING, DEV_STATE_ALARM,)

    # ------------------------------------------------------------------
    #   read & write AutoZero
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_AutoZero(self, the_att):
        """
        Method to read the AutoZero.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.auto_zero)

    @ExceptionHandler
    def write_AutoZero(self, the_att):
        """
        Method to set the AutoZero
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.auto_zero = data[0]
        self.push_change_event('AutoZero', data[0])

    def is_AutoZero_allowed(self, req_type):
        """
        Method to check if the AutoZero is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write AutoZeroEnable
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_AutoZeroEnable(self, the_att):
        """
        Method to read the AutoZeroEnable.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(eval(self.keithley.auto_zero_enable))

    @ExceptionHandler
    def write_AutoZeroEnable(self, the_att):
        """
        Method to set the AutoZeroEnable
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        if data[0]:
            self.keithley.auto_zero_enable = 'ON'
        if not data[0]:
            self.keithley.auto_zero_enable = 'OFF'
        self.push_change_event('AutoZeroEnable', data[0])

    def is_AutoZeroEnable_allowed(self, req_type):
        """
        Method to check if the AutoZeroEnable is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write AcquisitionMethod
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_AcquisitionMethod(self, the_att):
        """
        Method to read the AcquisitionMethod.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.acquisition_method)

    @ExceptionHandler
    def write_AcquisitionMethod(self, the_att):
        """
        Method to set the AcquisitionMethod
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.acquisition_method = data[0]
        self.push_change_event('AcquisitionMethod', data[0])

    def is_AcquisitionMethod_allowed(self, req_type):
        """
        Method to check if the AcquisitionMethod is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write LineSynchronization
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_LineSynchronization(self, the_att):
        """
        Method to read the LineSynchronization.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(eval(self.keithley.line_synchronization))

    @ExceptionHandler
    def write_LineSynchronization(self, the_att):
        """
        Method to set the LineSynchronization
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        if data[0]:
            self.keithley.line_synchronization = 'ON'
        if not data[0]:
            self.keithley.line_synchronization = 'OFF'
        self.push_change_event('LineSynchronization', data[0])


    def is_LineSynchronization_allowed(self, req_type):
        """
        Method to check if the LineSynchronization is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write InitiateContinuous
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_InitiateContinuous(self, the_att):
        """
        Method to read the InitiateContinuous.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(eval(self.keithley.initiate_continuous))
    @ExceptionHandler
    def write_InitiateContinuous(self, the_att):
        """
        Method to set the InitiateContinuous
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        if data[0]:
            self.keithley.initiate_continuous = 'ON'
        if not data[0]:
            self.keithley.initiate_continuous = 'OFF'
        self.push_change_event('InitiateContinuous', data[0])

    def is_InitiateContinuous_allowed(self, req_type):
        """
        Method to check if the InitiateContinuous is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write CountArmLayer
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_CountArmLayer(self, the_att):
        """
        Method to read the CountArmLayer.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(eval(self.keithley.count_arm_control))

    @ExceptionHandler
    def write_CountArmLayer(self, the_att):
        """
        Method to set the CountArmLayer
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.count_arm_control = data[0]
        self.push_change_event('CountArmLayer', data[0])

    def is_CountArmLayer_allowed(self, req_type):
        """
        Method to check if the CountArmLayer is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write CountScanLayer
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_CountScanLayer(self, the_att):
        """
        Method to read the CountScanLayer.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(eval(self.keithley.count_scan_control))

    @ExceptionHandler
    def write_CountScanLayer(self, the_att):
        """
        Method to set the CountScanLayer
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.count_scan_control = data[0]
        self.push_change_event('CountScanLayer', data[0])

    def is_CountScanLayer_allowed(self, req_type):
        """
        Method to check if the CountScanLayer is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write CountMeasureLayer
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_CountMeasureLayer(self, the_att):
        """
        Method to read the CountMeasureLayer.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(eval(self.keithley.count_measure_control))

    @ExceptionHandler
    def write_CountMeasureLayer(self, the_att):
        """
        Method to set the CountMeasureLayer
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.count_measure_control = data[0]
        self.push_change_event('CountMeasureLayer', data[0])

    def is_CountMeasureLayer_allowed(self, req_type):
        """
        Method to check if the CountMeasureLayer is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write DelayScanLayer
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_DelayScanLayer(self, the_att):
        """
        Method to read the DelayScanLayer.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.delay_scan_control)

    @ExceptionHandler
    def write_DelayScanLayer(self, the_att):
        """
        Method to set the DelayScanLayer
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.delay_scan_control = data[0]
        self.push_change_event('DelayScanLayer', data[0])

    def is_DelayScanLayer_allowed(self, req_type):
        """
        Method to check if the DelayScanLayer is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write DelayMeasureLayer
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_DelayMeasureLayer(self, the_att):
        """
        Method to read the DelayMeasureLayer.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.delay_measure_control)

    @ExceptionHandler
    def write_DelayMeasureLayer(self, the_att):
        """
        Method to set the DelayMeasureLayer
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.delay_measure_control = data[0]
        self.push_change_event('DelayMeasureLayer', data[0])

    def is_DelayMeasureLayer_allowed(self, req_type):
        """
        Method to check if the DelayMeasureLayer is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write SourceArmLayer
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_SourceArmLayer(self, the_att):
        """
        Method to read the SourceArmLayer.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.source_arm_control)

    @ExceptionHandler
    def write_SourceArmLayer(self, the_att):
        """
        Method to set the SourceArmLayer
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.source_arm_control = data[0]
        self.push_change_event('SourceArmLayer', data[0])

    def is_SourceArmLayer_allowed(self, req_type):
        """
        Method to check if the SourceArmLayer is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write SourceScanLayer
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_SourceScanLayer(self, the_att):
        """
        Method to read the SourceScanLayer.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.source_scan_control)

    @ExceptionHandler
    def write_SourceScanLayer(self, the_att):
        """
        Method to set the SourceScanLayer
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.source_scan_control = data[0]
        self.push_change_event('SourceScanLayer', data[0])

    def is_SourceScanLayer_allowed(self, req_type):
        """
        Method to check if the SourceScanLayer is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write SourceMeasureLayer
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_SourceMeasureLayer(self, the_att):
        """
        Method to read the SourceMeasureLayer.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.source_measure_control)

    @ExceptionHandler
    def write_SourceMeasureLayer(self, the_att):
        """
        Method to set the SourceMeasureLayer
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.source_measure_control = data[0]
        self.push_change_event('SourceMeasureLayer', data[0])

    def is_SourceMeasureLayer_allowed(self, req_type):
        """
        Method to check if the SourceMeasureLayer is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write TimerScanLayer
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_TimerScanLayer(self, the_att):
        """
        Method to read the TimerScanLayer.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.timer_scan_layer)

    @ExceptionHandler
    def write_TimerScanLayer(self, the_att):
        """
        Method to set the TimerScanLayer
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.timer_scan_layer = data[0]
        self.push_change_event('TimerScanLayer', data[0])

    def is_TimerScanLayer_allowed(self, req_type):
        """
        Method to check if the TimerScanLayer is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write TimerMeasureLayer
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_TimerMeasureLayer(self, the_att):
        """
        Method to read the TimerMeasureLayer.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.timer_measure_layer)

    @ExceptionHandler
    def write_TimerMeasureLayer(self, the_att):
        """
        Method to set the TimerMeasureLayer
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.timer_measure_layer = data[0]
        self.push_change_event('TimerMeasureLayer', data[0])

    def is_TimerMeasureLayer_allowed(self, req_type):
        """
        Method to check if the TimerMeasureLayer is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write CoupleTriggers
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_CoupleTriggers(self, the_att):
        """
        Method to read the CoupleTriggers.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.couple_triggers)

    @ExceptionHandler
    def write_CoupleTriggers(self, the_att):
        """
        Method to set the CoupleTriggers
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.couple_triggers = data[0]
        self.push_change_event('CoupleTriggers', data[0])

    def is_CoupleTriggers_allowed(self, req_type):
        """
        Method to check if the CoupleTriggers is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write InlineArmControl
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_InlineArmControl(self, the_att):
        """
        Method to read the InlineArmControl.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.inline_arm_control)

    @ExceptionHandler
    def write_InlineArmControl(self, the_att):
        """
        Method to set the InlineArmControl
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.inline_arm_control = data[0]
        self.push_change_event('InlineArmControl', data[0])

    def is_InlineArmControl_allowed(self, req_type):
        """
        Method to check if the InlineArmControl is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write InlineScanControl
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_InlineScanControl(self, the_att):
        """
        Method to read the InlineScanControl.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.inline_scan_control)

    @ExceptionHandler
    def write_InlineScanControl(self, the_att):
        """
        Method to set the InlineScanControl
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.inline_scan_control = data[0]
        self.push_change_event('InlineScanControl', data[0])

    def is_InlineScanControl_allowed(self, req_type):
        """
        Method to check if the InlineScanControl is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write InlineMeasureControl
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_InlineMeasureControl(self, the_att):
        """
        Method to read the InlineMeasureControl.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.inline_measure_control)

    @ExceptionHandler
    def write_InlineMeasureControl(self, the_att):
        """
        Method to set the InlineMeasureControl
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.inline_measure_control = data[0]
        self.push_change_event('InlineMeasureControl', data[0])

    def is_InlineMeasureControl_allowed(self, req_type):
        """
        Method to check if the InlineMeasureControl is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write OutlineArmControl
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_OutlineArmControl(self, the_att):
        """
        Method to read the OutlineArmControl.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.outline_arm_control)

    @ExceptionHandler
    def write_OutlineArmControl(self, the_att):
        """
        Method to set the OutlineArmControl
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.outline_arm_control = data[0]
        self.push_change_event('OutlineArmControl', data[0])

    def is_OutlineArmControl_allowed(self, req_type):
        """
        Method to check if the OutlineArmControl is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write OutlineScanControl
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_OutlineScanControl(self, the_att):
        """
        Method to read the OutlineScanControl.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.outline_scan_control)

    @ExceptionHandler
    def write_OutlineScanControl(self, the_att):
        """
        Method to set the OutlineScanControl
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.outline_scan_control = data[0]
        self.push_change_event('OutlineScanControl', data[0])

    def is_OutlineScanControl_allowed(self, req_type):
        """
        Method to check if the OutlineScanControl is available to interact with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write OutlineMeasureControl
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_OutlineMeasureControl(self, the_att):
        """
        Method to read the OutlineMeasureControl.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.outline_measure_control)

    @ExceptionHandler
    def write_OutlineMeasureControl(self, the_att):
        """
        Method to set the OutlineMeasureControl
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.outline_measure_control = data[0]
        self.push_change_event('OutlineMeasureControl', data[0])

    def is_OutlineMeasureControl_allowed(self, req_type):
        """
        Method to check if the OutlineMeasureControl is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write SemiSynchronousLine
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_SemiSynchronousLine(self, the_att):
        """
        Method to read the SemiSynchronousLine.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.semisynchronous_line)

    @ExceptionHandler
    def write_SemiSynchronousLine(self, the_att):
        """
        Method to set the SemiSynchronousLine
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.semisynchronous_line = data[0]
        self.push_change_event('SemiSynchronousLine', data[0])

    def is_SemiSynchronousLine_allowed(self, req_type):
        """
        Method to check if the SemiSynchronousLine is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write EventEnableRegister
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_EventEnableRegister(self, the_att):
        """
        Method to read the EventEnableRegister.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.event_enable_register)

    @ExceptionHandler
    def write_EventEnableRegister(self, the_att):
        """
        Method to set the EventEnableRegister
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.event_enable_register = data[0]
        self.push_change_event('EventEnableRegister', data[0])

    def is_EventEnableRegister_allowed(self, req_type):
        """
        Method to check if the EventEnableRegister is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_RUNNING,
                                    DEV_STATE_FAULT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read EventStatusRegister
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_EventStatusRegister(self, the_att):
        """
        Method to read the EventStatusRegister.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.event_status_register)

    def is_EventStatusRegister_allowed(self, req_type):
        """
        Method to check if the EventStatusRegister is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_RUNNING,
                                    DEV_STATE_FAULT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read Identification
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_Identification(self, the_att):
        """
        Method to read the Identification.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.identification)

    def is_Identification_allowed(self, req_type):
        """
        Method to check if the Identification is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write OperationCompleted
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_OperationCompleted(self, the_att):
        """
        Method to read the OperationCompleted.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.operation_complete)

    @ExceptionHandler
    def write_OperationCompleted(self, the_att):
        """
        Method to set the OperationCompleted
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.operation_complete = data[0]
        self.push_change_event('OperationCompleted', data[0])

    def is_OperationCompleted_allowed(self, req_type):
        """
        Method to check if the OperationCompleted is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_RUNNING,
                                    DEV_STATE_FAULT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read MemoryOptionInstalled
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_MemoryOptionInstalled(self, the_att):
        """
        Method to read the MemoryOptionInstalled.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.option_identification)

    def is_MemoryOptionInstalled_allowed(self, req_type):
        """
        Method to check if the MemoryOptionInstalled is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write ServiceRequestEnableRegister
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_ServiceRequestEnableRegister(self, the_att):
        """
        Method to read the ServiceRequestEnableRegister.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.service_request_enable_register)

    @ExceptionHandler
    def write_ServiceRequestEnableRegister(self, the_att):
        """
        Method to set the ServiceRequestEnableRegister
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.service_request_enable_register = data[0]
        self.push_change_event('ServiceRequestEnableRegister', data[0])

    def is_ServiceRequestEnableRegister_allowed(self, req_type):
        """
        Method to check if the ServiceRequestEnableRegister is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_RUNNING,
                                    DEV_STATE_FAULT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write StatusByteRegister
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_StatusByteRegister(self, the_att):
        """
        Method to read the StatusByteRegister.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(eval(self.keithley.status_byte_register))

    @ExceptionHandler
    def write_StatusByteRegister(self, the_att):
        """
        Method to set the StatusByteRegister
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.status_byte_register = data[0]
        self.push_change_event('StatusByteRegister', data[0])

    def is_StatusByteRegister_allowed(self, req_type):
        """
        Method to check if the StatusByteRegister is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_RUNNING,
                                    DEV_STATE_FAULT, DEV_STATE_ON,  )

    # ------------------------------------------------------------------
    #   read ErrorFlag
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_ErrorFlag(self, the_att):
        """
        Method to read the ErrorFlag.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.error_flag)

    def is_ErrorFlag_allowed(self, req_type):
        """
        Method to check if the ErrorFlag is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_RUNNING,
                                    DEV_STATE_FAULT, DEV_STATE_ON,
                                    DEV_STATE_MOVING )

    # ------------------------------------------------------------------
    #   read ReadNonBlocking
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_ReadNonBlocking(self, the_att):
        """
        Method to read the ReadNonBlocking.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.read_nonblocking)

    def is_ReadNonBlocking_allowed(self, req_type):
        """
        Method to check if the ReadNonBlocking is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_RUNNING, )

    # ------------------------------------------------------------------
    #   read RawData
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_RawData(self, the_att):
        """
        Method to read the RawData.
        :param the_att: actual name of the attribute
        :return: None
        """

        if self.raw_data == []:
            the_att.set_value('No data')
        else:
            the_att.set_value(self.raw_data[-1])

    def is_RawData_allowed(self, req_type):
        """
        Method to check if the RawData is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_RUNNING, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read RawDataBuffer
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_RawDataBuffer(self, the_att):
        """
        Method to read the RawData.
        :param the_att: actual name of the attribute
        :return: None
        """
        print('DEBUG: ' + str(self.raw_data))
        the_att.set_value(self.raw_data)
        # if self.raw_data == []:
        #     the_att.set_value('No data')
        # else:
        #     the_att.set_value(self.raw_data)

    def is_RawDataBuffer_allowed(self, req_type):
        """
        Method to check if the RawData is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_RUNNING, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write AcqModeEnable
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_AcqModeEnable(self, the_att):
        """
        Method to read the AcqModeEnable.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.acq_mode)

    @ExceptionHandler
    def write_AcqModeEnable(self, the_att):
        """
        Method to set the AcqModeEnable
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.acq_mode = data[0]
        self.push_change_event('AcqModeEnable', data[0])

    def is_AcqModeEnable_allowed(self, req_type):
        """
        Method to check if the AcqModeEnable is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write TimerPeriod
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_TimerPeriod(self, the_att):
        """
        Method to read the TimerPeriod.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.timer_measure_layer)

    @ExceptionHandler
    def write_TimerPeriod(self, the_att):
        """
        Method to set the TimerPeriod
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.timer_measure_layer = data[0]
        self.push_change_event('TimerPeriod', data[0])

    def is_TimerPeriod_allowed(self, req_type):
        """
        Method to check if the TimerPeriod is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read TriggerEventRegister
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_TriggerEventRegister(self, the_att):
        """
        Method to read the TriggerEventRegister.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.trigger_event_register)

    def is_TriggerEventRegister_allowed(self, req_type):
        """
        Method to check if the TriggerEventRegister is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON,
                                    DEV_STATE_RUNNING, )

    # ------------------------------------------------------------------
    #   read OperationEventRegister
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_OperationEventRegister(self, the_att):
        """
        Method to read the OperationEventRegister.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.operation_event_register)

    def is_OperationEventRegister_allowed(self, req_type):
        """
        Method to check if the OperationEventRegister is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON,
                                    DEV_STATE_RUNNING, )

    # ------------------------------------------------------------------
    #   read & write ChunkSize
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_ChunkSize(self, the_att):
        """
        Method to read the OperationEventRegister.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.data_ready_chunk)

    @ExceptionHandler
    def write_ChunkSize(self, the_att):
        """
        Method to set the TimerPeriod
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.data_ready_chunk = data[0]
        self.push_change_event('ChunkSize', data[0])

    def is_ChunkSize_allowed(self, req_type):
        """
        Method to check if the OperationEventRegister is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON,
                                    DEV_STATE_RUNNING, )

    # ------------------------------------------------------------------
    #   read DataReady
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_DataReady(self, the_att):
        """
        Method to read the OperationEventRegister.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(len(self.raw_data))

    def is_DataReady_allowed(self, req_type):
        """
        Method to check if the OperationEventRegister is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON,
                                    DEV_STATE_RUNNING, )

    # ------------------------------------------------------------------
    #   read & write BufferElementGroup
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_BufferElementGroup(self, the_att):
        """
        Method to read the BufferElementGroup.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.buffer_elem_group)

    @ExceptionHandler
    def write_BufferElementGroup(self, the_att):
        """
        Method to set the TimerPeriod
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.buffer_elem_group = data[0]
        self.push_change_event('BufferElementGroup', data[0])

    def is_BufferElementGroup_allowed(self, req_type):
        """
        Method to check if the OperationEventRegister is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write BufferSize
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_BufferSize(self, the_att):
        """
        Method to read the BufferSize.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.buffer_size)

    @ExceptionHandler
    def write_BufferSize(self, the_att):
        """
        Method to set the TimerPeriod
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.buffer_size = data[0]
        self.push_change_event('BufferSize', data[0])

    def is_BufferSize_allowed(self, req_type):
        """
        Method to check if the OperationEventRegister is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write BufferSizeAuto
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_BufferSizeAuto(self, the_att):
        """
        Method to read the BufferSizeAuto.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(eval(self.keithley.buffer_size_auto))

    @ExceptionHandler
    def write_BufferSizeAuto(self, the_att):
        """
        Method to set the TimerPeriod
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        if data[0]:
            self.keithley.buffer_size_auto = 'ON'
        if not data[0]:
            self.keithley.buffer_size_auto = 'OFF'
        self.push_change_event('BufferSizeAuto', data[0])

    def is_BufferSizeAuto_allowed(self, req_type):
        """
        Method to check if the OperationEventRegister is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON, )

    # ------------------------------------------------------------------
    #   read & write BufferFeed
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_BufferFeed(self, the_att):
        """
        Method to read the BufferFeed.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.buffer_feed)

    @ExceptionHandler
    def write_BufferFeed(self, the_att):
        """
        Method to set the TimerPeriod
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.buffer_feed = data[0]
        self.push_change_event('BufferFeed', data[0])

    def is_BufferFeed_allowed(self, req_type):
        """
        Method to check if the OperationEventRegister is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON,)

    # ------------------------------------------------------------------
    #   read & write BufferControl
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_BufferControl(self, the_att):
        """
        Method to read the BufferControl.
        :param the_att: actual name of the attribute
        :return: None
        """
        the_att.set_value(self.keithley.buffer_control)

    @ExceptionHandler
    def write_BufferControl(self, the_att):
        """
        Method to set the TimerPeriod
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        the_att.get_write_value(data)
        self.keithley.buffer_control = data[0]
        self.push_change_event('BufferControl', data[0])

    def is_BufferControl_allowed(self, req_type):
        """
        Method to check if the OperationEventRegister is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON,)

    # ------------------------------------------------------------------
    #   read BufferData
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_BufferData(self, the_att):
        """
        Method to read the BufferData.
        :param the_att: actual name of the attribute
        :return: None
        """
        data = []
        data = self.keithley.buffer_data.split(',')
        print(data)
        the_att.set_value(data)

    def is_BufferData_allowed(self, req_type):
        """
        Method to check if the OperationEventRegister is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON,)

    # ------------------------------------------------------------------
    #   read BufferDataStored
    # ------------------------------------------------------------------
    @ExceptionHandler
    def read_BufferDataStored(self, the_att):
        """
        Method to read the BufferDataStored.
        :param the_att: actual name of the attribute
        :return: None
        """
        data_available, data_used = self.keithley.buffer_free.split(',')
        data = eval(data_used)/4
        the_att.set_value(data)

    def is_BufferDataStored_allowed(self, req_type):
        """
        Method to check if the OperationEventRegister is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON,)


    # ------------------------------------------------------------------
    #   COMMANDS
    # ------------------------------------------------------------------

    # @ExceptionHandler
    # def FunctionMeasurement(self, function):
    #     """
    #     Cmd to perform a measurement with the input function
    #     :param the_att: actual name of the attribute
    #     :return: None
    #     """
    #     return self.keithley.get_measure(function)
    #
    # def is_FunctionMeasurement_allowed(self):
    #     """
    #     Method to check if performing a measurement is allowed.
    #     :return: None
    #     """
    #     return self.get_state() in (DEV_STATE_INIT, DEV_STATE_RUNNING,
    #                                 DEV_STATE_FAULT, DEV_STATE_ON, )

    # @ExceptionHandler
    # def TriggerBus(self):
    #     """
    #     Method that triggers the gpib bus
    #     :return:
    #     """
    #     self.keithley.trigger()
    #
    # def is_TriggerBus_allowed(self):
    #     """
    #     Method to check if triggering the HW is allowed.
    #     :return:
    #     """
    #     return self.get_state() in (DEV_STATE_INIT, DEV_STATE_RUNNING,
    #                                 DEV_STATE_FAULT, DEV_STATE_ON, )

    @ExceptionHandler
    def ClearStatus(self):
        """
        Method that Clears all event registers, and Error Queue.
        :return:
        """
        self.keithley.clear_status()

    def is_ClearStatus_allowed(self):
        """
        Method to check if CLS command is allowed.
        :return:
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_FAULT,
                                    DEV_STATE_ON, )

    @ExceptionHandler
    def Reset(self):
        """
        Method that performs a HW reset cmd
        :return:
        """
        self.keithley.reset()

    def is_Reset_allowed(self):
        """
        Method to check if the command is allowed.
        :return:
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_FAULT,
                                    DEV_STATE_ON, )

    @ExceptionHandler
    def SaveSetup(self):
        """
        Method that performs a HW save cmd
        :return:
        """
        self.keithley.sense_1_save_alternate()

    def is_SaveSetup_allowed(self):
        """
        Method to check if the command is allowed.
        :return:
        """
        return self.get_state() in (DEV_STATE_ON, DEV_STATE_INIT, )

    @ExceptionHandler
    def LoadSetup(self):
        """
        Method that performs a HW load cmd
        :return:
        """
        self.keithley.sense_1_load_alternate()

    def is_LoadSetup_allowed(self):
        """
        Method to check if the command is allowed.
        :return:
        """
        return self.get_state() in (DEV_STATE_ON, DEV_STATE_INIT, )

    # @ExceptionHandler
    # def CurrentAcquire(self):
    #     """
    #     Method that performs a HW current acquisition
    #     :return:
    #     """
    #     return self.keithley.get_measure('CURR')
    #
    # def is_CurrentAcquire_allowed(self):
    #     """
    #     Method to check if the command is allowed.
    #     :return:
    #     """
    #     return self.get_state() in (DEV_STATE_INIT, DEV_STATE_RUNNING,
    #                                 DEV_STATE_FAULT, DEV_STATE_ON, )

    # @ExceptionHandler
    # def VoltageAcquire(self):
    #     """
    #     Method that performs a HW current acquisition
    #     :return:
    #     """
    #     return self.keithley.get_measure('VOLT')
    #
    # def is_VoltageAcquire_allowed(self):
    #     """
    #     Method to check if the command is allowed.
    #     :return:
    #     """
    #     return self.get_state() in (DEV_STATE_INIT, DEV_STATE_RUNNING,
    #                                 DEV_STATE_FAULT, DEV_STATE_ON, )

    @ExceptionHandler
    def Preset(self):
        """
        Method that performs a HW preset
        :return:
        """
        self.keithley.preset()

    def is_Preset_allowed(self):
        """
        Method to check if the command is allowed.
        :return:
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_FAULT,
                                    DEV_STATE_ON, )

    @ExceptionHandler
    def ClearErrorQueue(self):
        """
        Method that performs a HW clear of the error queue
        :return:
        """
        self.keithley.error_msg_clear()
        self.error_list = []

    def is_ClearErrorQueue_allowed(self):
        """
        Method to check if the command is allowed.
        :return:
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_FAULT,
                                    DEV_STATE_ON, )

    @ExceptionHandler
    def Start(self):
        """
        Method that performs a HW initiate
        :return:
        """
        self.idle = False
        # self.idle = not self.keithley.trigger_event_register & 2**1 != 0
        if not self.idle:
            self.set_state(DEV_STATE_RUNNING)
            if self.keithley.source_measure_control == 'IMM':
                self.keithley.count_measure_control = 1
                self.set_status('Keithley on measure state.')
                self.push_change_event('State', self.get_state())
                self.push_change_event('Status', self.get_status())
                self.keithley.initiate()
                self.raw_data += [self.keithley.read_latest]
                # self.push_data_ready_event('DataReady', 1)

            elif self.keithley.source_measure_control in ['EXT', 'TLIN', 'TIM']:
                self.quit_event.clear()
                self.set_status('Keithley on acquiring mode.')
                self.push_change_event('State', self.get_state())
                self.push_change_event('Status', self.get_status())
                self.async = True
                t = threading.Thread(target=self._acquisition_thread,
                                     args=[self.quit_event])
                t.start()

    def is_Start_allowed(self):
        """
        Method to check if the command is allowed.
        :return:
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_ON)

    def _acquisition_thread(self, quit_event):
        """
        Method executed by the thread in charge of retrieving readings from the
        device asynchronously
        :return: read value
        """
        self.quit_event.clear()
        count = eval(self.keithley.count_measure_control)
        self.count_index = 0
        self.keithley.initiate()
        while not self.idle and not quit_event.isSet() and not \
                self.count_index == count:
            if not quit_event.isSet():
                self.raw_data += [self.keithley.read_fresh]
                self.count_index += 1
                print(self.count_index)
                if self.count_index % self.data_ready_chunk == 0:
                    self._generate_data_ready_event()
        # Force last update
        if not quit_event.isSet() and not self.count_index % \
                self.data_ready_chunk == 0:
            self._generate_data_ready_event()

        self.async = False

    def _generate_data_ready_event(self):
        try:
            self.push_data_ready_event('DataReady', len(self.raw_data))
            print('++++++++Event sent.')
        except Exception as e:
            print('\n\n\n\n\n\n ----------------------Event generated lost: '
                  '%s' % e)

    @ExceptionHandler
    def Stop(self):
        """
        Method that performs a HW abort
        :return:
        """
        self.quit_event.set()

        self.keithley.abort()

        self.idle = True
        self.set_state(DEV_STATE_ON)
        self.set_status('Keithley on idle state.')
        self.push_change_event('State', self.get_state())
        self.push_change_event('Status', self.get_status())

    def is_Stop_allowed(self):
        """
        Method to check if the command is allowed.
        :return:
        """
        return self.get_state() in (DEV_STATE_FAULT, DEV_STATE_RUNNING,
                                    DEV_STATE_ON,)

    @ExceptionHandler
    def Wait(self):
        """
        Method that performs a HW wait
        :return:
        """
        self.keithley.wait()

    def is_wait_allowed(self):
        """
        Method to check if the command is allowed.
        :return:
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_RUNNING,
                                    DEV_STATE_FAULT, DEV_STATE_ON, )

    @ExceptionHandler
    def NonBlockingReadConfig(self):
        """
        Method that performs a HW configuration so non reading can be performed
        :return:
        """
        self.keithley.hardware_nonblocking_read_config()

    def is_NonBlockingReadConfig_allowed(self):
        """
        Method to check if the command is allowed.
        :return:
        """
        return self.get_state() in (DEV_STATE_ON, DEV_STATE_INIT, )

        # TODO: set internal flags (searching, fault, etc) when necessary

    @ExceptionHandler
    def MakeAcquisition(self, points_to_scan):
        """
        Cmd to perform a measurement of points_to_scan points
        :return: None
        """
        return self.keithley.acquisition(points_to_scan)

    def is_MakeAcquisition_allowed(self):
        """
        Method to check if performing a measurement is allowed.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_RUNNING,
                                    DEV_STATE_FAULT, DEV_STATE_ON, )

    @ExceptionHandler
    def PrepareAcquisition(self):
        """
        Cmd to prepare the config for making a measurement
        points
        :return: None
        """
        self.keithley.hardware_nonblocking_read_config()
        self.raw_data = []
        self.error_list = []
        self.acq_mode = 1
        self.count_index = 0

    def is_PrepareAcquisition_allowed(self):
        """
        Method to check if performing a measurement is allowed.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_FAULT,
                                    DEV_STATE_ON,)

    @ExceptionHandler
    def ClearBuffer(self):
        """
        Cmd to prepare the config for making a measurement
        points
        :return: None
        """
        self.keithley.buffer_clear()

    def is_ClearBuffer_allowed(self):
        """
        Method to check if performing a measurement is allowed.
        :return: None
        """
        return self.get_state() in (DEV_STATE_INIT, DEV_STATE_FAULT,
                                    DEV_STATE_ON,)

    @ExceptionHandler
    def RawDataIndex(self, params):
        """
        Method to read a RawData buffer chunk.
        :param params: list with pointers to the first and end position
        :return:
        """
        start_index, end_index = params
        print(params)
        chunk = self.raw_data[start_index:end_index]
        return np.array(chunk)


    def is_RawDataIndex_allowed(self):
        """
        Method to check if the RawDataIndex is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_RUNNING, DEV_STATE_ON, )
    #
    # @ExceptionHandler
    # def RawCmdRead(self):
    #     """
    #     Method to read a RawData buffer chunk.
    #     :param params: list with pointers to the first and end position
    #     :return:
    #     """
    #     return self.keithley.read_cmd()
    #
    # def is_RawCmdRead_allowed(self):
    #     """
    #     Method to check if the RawCmdRead is available to interact
    #     with.
    #     :return: None
    #     """
    #     return self.get_state() in (DEV_STATE_RUNNING, DEV_STATE_ON,
    #                                 DEV_STATE_INIT, DEV_STATE_FAULT,)

    @ExceptionHandler
    def RawCmdWrite(self, params):
        """
        Method to read a RawData buffer chunk.
        :param params: list with pointers to the first and end position
        :return:
        """
        self.keithley.write_cmd(params)

    def is_RawCmdWrite_allowed(self):
        """
        Method to check if the RawCmdWrite is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_RUNNING, DEV_STATE_ON,
                                    DEV_STATE_INIT, DEV_STATE_FAULT,)

    @ExceptionHandler
    def RawCmdWriteRead(self, params):
        """
        Method to read a RawData buffer chunk.
        :param params: list with pointers to the first and end position
        :return:
        """
        self.keithley.write_cmd(params)
        return self.keithley.read_cmd()

    def is_RawCmdWriteRead_allowed(self):
        """
        Method to check if the RawCmdWriteRead is available to interact
        with.
        :return: None
        """
        return self.get_state() in (DEV_STATE_RUNNING, DEV_STATE_ON,
                                    DEV_STATE_INIT, DEV_STATE_FAULT,)

    def __del__(self):
        """
        Method to close the socket when deleting the object.
        """
        self.keithley.abort()
